from django.contrib.auth.models import User
from django.contrib.sessions.models import Session
from django.db import models


class Timestamped(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class KeycloakUser(Timestamped):
    """
    This model represents a Keycloak user and associates them with a corresponding Django user.
    """

    sub = models.UUIDField(primary_key=True, db_index=True)
    user = models.OneToOneField(User, on_delete=models.CASCADE)

    def __str__(self):
        return f"Keycloak user id: {self.sub} linked to Django user: {self.user}"


class KeycloakSession(models.Model):
    """
    This model is used to track and combine created Keycloak and Django sessions together.
    """

    keycloak_session_id = models.UUIDField(editable=False, db_index=True)
    keycloak_user = models.ForeignKey(KeycloakUser, on_delete=models.CASCADE)
    django_session = models.OneToOneField(
        Session, on_delete=models.CASCADE, editable="False", unique=True
    )
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"Keycloak user id: {self.keycloak_user_id} linked to Django session: {self.django_session.session_key} and Keycloak session: {self.keycloak_session_id}"



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Add custom permissions"

    def handle(self, *args, **kwargs):
        content_type = ContentType.objects.get_for_model(Permission)
        custom_permissions = [
            {"codename": "view_dashboard", "name": "Can view dashboard"},
        ]

        for perm in custom_permissions:
            _, created = Permission.objects.get_or_create(
                codename=perm["codename"],
                name=perm["name"],
                content_type=content_type,
            )
            if created:
                self.stdout.write(
                    self.style.SUCCESS(f"Permission '{perm['name']}' created.")
                )
            else:
                self.stdout.write(f"Permission '{perm['name']}' already exists.")



from django.conf import settings
from keycloak import KeycloakOpenID, KeycloakOpenIDConnection
from keycloak.exceptions import (
    KeycloakDeleteError,
    KeycloakGetError,
    raise_error_from_response,
)

BACKCHANNEL_LOGOUT_EVENT_URL = "http://schemas.openid.net/event/backchannel-logout"
BACKCHANNEL_LOGOUT_EVENT_HTTPS_URL = (
    "https://schemas.openid.net/event/backchannel-logout"
)
REALM_URL = f"{settings.KC_SERVER_URL}/realms/{settings.KC_REALM}"


keycloak_connection = KeycloakOpenIDConnection(
    server_url=settings.KC_SERVER_URL,
    realm_name=settings.KC_REALM,
    user_realm_name="master",
    username="admin",
    password="admin",
    verify=True,
)

keycloak_openid = KeycloakOpenID(
    server_url=settings.KC_SERVER_URL,
    realm_name=settings.KC_REALM,
    client_id=settings.KC_CLIENT_ID,
    client_secret_key=settings.KC_CLIENT_SECRET,
)


def delete_session(session_id):
    path = "admin/realms/{realm-name}/sessions/{session_id}"
    params_path = {
        "realm-name": keycloak_connection.realm_name,
        "session_id": session_id,
    }
    data_raw = keycloak_connection.raw_delete(path.format(**params_path))
    return raise_error_from_response(data_raw, KeycloakDeleteError)


def get_logout_url(id_token, post_logout_redirect_uri):
    return (
        f"{REALM_URL}/protocol/openid-connect/logout"
        f"?id_token_hint={id_token}"
        f"&post_logout_redirect_uri={post_logout_redirect_uri}"
        f"&client_id={settings.KC_CLIENT_ID}"
    )


def get_active_devices(access_token):
    connection = keycloak_openid.connection
    path = "realms/{realm-name}/account/sessions/devices"
    orig_bearer = connection.headers.get("Authorization")
    connection.add_param_headers("Authorization", "Bearer " + access_token)
    connection.add_param_headers("Content-Type", "application/json")

    params_path = {"realm-name": keycloak_openid.realm_name}
    data_raw = connection.raw_get(path.format(**params_path))
    (
        connection.add_param_headers("Authorization", orig_bearer)
        if orig_bearer is not None
        else connection.del_param_headers("Authorization")
    )
    return raise_error_from_response(data_raw, KeycloakGetError)



from django.conf import settings
from django.urls import path

from . import views

urlpatterns = [
    path(
        getattr(settings, "KC_LOGIN_URL", "login/"),
        views.LoginView.as_view(),
        name="kc_auth_login",
    ),
    path(
        getattr(settings, "KC_CALLBACK_URL", "callback/"),
        views.CallbackView.as_view(),
        name="kc_auth_callback",
    ),
    path(
        getattr(settings, "KC_LOGOUT_URL", "logout/"),
        views.LogoutView.as_view(),
        name="kc_auth_logout",
    ),
    path(
        getattr(settings, "KC_REMOTE_LOGOUT_URL", "remote-logout/"),
        views.RemoteLogoutView.as_view(),
        name="kc_auth_remote-logout",
    ),
    path(
        getattr(settings, "KC_LOGOUT_LISTENER_URL", "logout-listener/"),
        views.LogoutListenerView.as_view(),
        name="kc_auth_logout-listener",
    ),
    path(
        getattr(settings, "KC_DEVICES_URL", "devices/"),
        views.devices,
        name="kc_auth_devices",
    ),
    path(
        getattr(settings, "KC_DEVICES_API_URL", "api/devices/"),
        views.DevicesAPIView.as_view(),
        name="kc_auth_api_devices",
    ),
]



import json
import logging

import jwt
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.sessions.models import Session
from django.http import Http404, JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from keycloak import KeycloakError
from rest_framework import permissions
from rest_framework.views import APIView

from .keycloak_openid_config import (
    BACKCHANNEL_LOGOUT_EVENT_HTTPS_URL,
    BACKCHANNEL_LOGOUT_EVENT_URL,
    REALM_URL,
    delete_session,
    get_active_devices,
    get_logout_url,
    keycloak_openid,
)
from .models import KeycloakSession, KeycloakUser

logger = logging.getLogger(__name__)


class LoginView(View):
    def get(self, request):
        """
        Handles user login by redirecting to the Keycloak authentication URL.

        The method constructs the Keycloak authentication URL with a redirect URI and state.
        The `state` parameter encodes the next URL the user should be redirected to after authentication.

        Parameters:
            request: The HTTP request object.

        Returns:
            HttpResponseRedirect: A response redirecting the user to the Keycloak authentication URL.
        """
        next_url = request.GET.get(
            "next", getattr(settings, "KC_SUCCESSFUL_LOGIN_REDIRECT", "/")
        )
        server_url = request.build_absolute_uri("/")[:-1]
        callback_url = f"{server_url}{reverse("kc_auth_callback")}"

        auth_url = keycloak_openid.auth_url(
            redirect_uri=callback_url,
            scope="openid email",
            state=urlsafe_base64_encode(next_url.encode("utf-8")),
        )
        logger.info("Redirecting user to Keycloak authentication URL")
        return redirect(auth_url)


class CallbackView(View):
    def get(self, request):
        """
        Processes the Keycloak callback after authentication.

        Parameters:
            request: The HTTP request object.

        Workflow:
            - Extracts `code` and `state` parameters from the request.
            - Decodes `state` to determine the redirection URL after login.
            - If no `code` is provided, redirects to the home or the state-defined URL.
            - Exchanges the authorization code for tokens via Keycloak.
            - Retrieves user information from Keycloak.
            - Authenticates and logs in the user.
            - Creates a Keycloak session record linked to the Django session.
            - Redirects the user to the state-defined URL or the home page.

        Returns:
            HttpResponseRedirect: Redirects to the appropriate URL based on the state or home page.
        """
        code = request.GET.get("code")
        state = request.GET.get("state")
        redirect_url = None

        if state:
            try:
                redirect_url = urlsafe_base64_decode(state).decode("ascii")
            except Exception:
                redirect_url = (
                    getattr(settings, "KC_SUCCESSFUL_LOGIN_REDIRECT", "home"),
                )
        print(redirect_url)
        redirect_url = redirect_url if redirect_url else "home"
        if not code:
            return redirect(redirect_url)

        callback_url = request.build_absolute_uri(request.path)

        try:
            token = keycloak_openid.token(
                grant_type="authorization_code",
                code=code,
                redirect_uri=callback_url,
            )
            id_token = token.get("id_token")
            access_token = token.get("access_token")
            keycloak_session_id = token.get("session_state")

            user_info = keycloak_openid.userinfo(access_token)
        except KeycloakError:
            redirect_error_message = getattr(settings, "KC_ERROR_MESSAGES", {}).get(
                "redirect_error"
            )
            logger.error("Error redirecting user %s", callback_url)
            if redirect_error_message:
                messages.error(
                    request,
                    redirect_error_message,
                )
            return redirect(redirect_url)

        user = authenticate(request, user_info=user_info)
        if not user:
            login_failed_message = getattr(settings, "KC_ERROR_MESSAGES", {}).get(
                "login_failed"
            )
            logger.error("Login in Django failed for user %s", user_info.get("sub"))
            if login_failed_message:
                messages.error(
                    request,
                    login_failed_message,
                )
            return redirect(redirect_url)

        login(request, user)
        logger.info("User %s logged in successfully", user.username)

        request.session["id_token"] = id_token
        request.session["access_token"] = access_token

        try:
            KeycloakSession.objects.get_or_create(
                keycloak_session_id=keycloak_session_id,
                keycloak_user=KeycloakUser.objects.get(user=user),
                django_session=Session.objects.get(
                    session_key=request.session.session_key
                ),
            )
        except Exception as e:
            logger.error("Failed to create Keycloak session record: %s", e)

        return redirect(redirect_url)


class LogoutView(LoginRequiredMixin, View):
    def post(self, request):
        """
        Handles the logout process by invalidating the Django session and redirecting
        to the Keycloak logout URL or the application homepage.

        Parameters:
            request: The HTTP request object.

        Returns:
            HttpResponseRedirect: Redirects to the Keycloak logout URL or the homepage.
        """
        server_url = request.build_absolute_uri("/")[:-1]
        post_logout_redirect_uri = (
            f"{server_url}{getattr(settings, "KC_LOGOUT_REDIRECT", "/")}"
        )
        id_token = request.session.get("id_token")

        user = request.user.username
        logout(request)
        logger.info("%s logged out", user)

        if not id_token:
            return redirect(post_logout_redirect_uri)

        redirect_url = get_logout_url(id_token, post_logout_redirect_uri)
        logger.info("Redirecting %s to Keycloak logout URL", user)

        return redirect(redirect_url)


class RemoteLogoutView(LoginRequiredMixin, View):
    def post(self, request):
        """
        Invalidates a remote Keycloak session associated with the authenticated user.

        Parameters:
            request: The HTTP request object.

        Returns:
            HttpResponseRedirect: Redirects to the referring page or home on failure.
        """
        session_id = request.POST.get("session_id") or json.loads(request.body).get(
            "session_id"
        )
        keycloak_user = KeycloakUser.objects.filter(user=request.user).first()
        if not keycloak_user:
            logger.error(
                "User with %s session not found.",
                session_id,
            )
            user_not_found_message = getattr(settings, "KC_ERROR_MESSAGES", {}).get(
                "user_not_found"
            )
            if user_not_found_message:
                messages.error(
                    request,
                    user_not_found_message,
                )
            return redirect(request.META.get("HTTP_REFERER", "/"))
        try:
            if keycloak_user.user == request.user:
                delete_session(session_id)
                logger.info(
                    "Successfully deleted Keycloak session: %s for user: %s",
                    session_id,
                    request.user.username,
                )
            else:
                raise KeycloakError("Users mismatch!")
        except KeycloakError as e:
            logger.error(
                "Failed to delete Keycloak session: %s for user: %s. Error: %s",
                session_id,
                request.user.username,
                str(e),
            )
            remote_logout_failed_message = getattr(
                settings, "KC_ERROR_MESSAGES", {}
            ).get("remote_logout_failed")
            if remote_logout_failed_message:
                messages.error(
                    request,
                    remote_logout_failed_message,
                )

        return redirect(request.META.get("HTTP_REFERER", "/"))


@method_decorator(csrf_exempt, name="dispatch")
class LogoutListenerView(View):
    def post(self, request):
        """
        Listens for Keycloak backchannel logout events and invalidates corresponding Django sessions.

        Parameters:
            request: The HTTP request object.

        Returns:
            JsonResponse: A success or failure response based on the processing result.
        """
        instance = jwt.JWT()
        logout_token = request.body.decode().split("=")[-1]
        logger.info("Received logout token from Keycloak.")

        try:
            logout_token = instance.decode(
                logout_token, settings.KC_VERIFYING_KEY, do_time_check=True
            )
            events = logout_token.get("events", {})
            iss = logout_token.get("iss", "")
            if (
                BACKCHANNEL_LOGOUT_EVENT_URL in events
                or BACKCHANNEL_LOGOUT_EVENT_HTTPS_URL in events
            ) and REALM_URL == iss:

                sid = logout_token.get("sid")

                session_keys = KeycloakSession.objects.filter(
                    keycloak_session_id=sid
                ).values_list("django_session", flat=True)

                deleted_sessions = Session.objects.filter(
                    session_key__in=session_keys
                ).delete()
                logger.info(
                    "Deleted %d Django session(s) for Keycloak session ID: %s",
                    deleted_sessions[0],
                    sid,
                )

                return JsonResponse({"status": "success"})
            else:
                logger.warning("Invalid logout event or issuer mismatch")
                return JsonResponse(
                    {"status": "failure", "reason": "Invalid logout event"}, status=400
                )
        except Exception as e:
            logger.error("Unexpected error processing logout token: %s", str(e))
            return JsonResponse(
                {"status": "failure", "reason": "Unexpected error"}, status=500
            )


@login_required
def devices(request):
    """
    Fetches active sessions and devices for the logged-in user from Keycloak.

    Parameters:
        request: The HTTP request object.

    Returns:
        HttpResponse: Renders a template with session and device information.

    Raises:
        Http404: If the access token is not found in the user's session.
    """
    template_name = "devices.html"
    session_data = []
    error = False
    access_token = request.session.get("access_token")
    if not access_token:
        logger.warning("Access token not found for user: %s", request.user.username)
        raise Http404
    try:
        session_data = get_active_devices(access_token)
        logger.info("Retrieved devices for user: %s", request.user.username)
    except KeycloakError as e:
        logger.error(
            "Failed to fetch active devices for user %s: %s",
            request.user.username,
            str(e),
        )
        error = True
    return render(
        request, template_name, {"session_data": session_data, "error": error}
    )


class DevicesAPIView(APIView):
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser)

    def get(self, request):
        """
        API endpoint that returns active sessions and devices for the logged-in user from Keycloak.

        Parameters:
            request: The HTTP request object.

        Returns:
            JsonResponse: JSON response containing session data or error information.
            - On success: Returns a JSON object with 'session_data' containing device information.
            - On error: Returns a JSON object with 'error' set to true and 'message' explaining the issue.

        HTTP Status Codes:
            200: Successfully retrieved device information.
            401: Access token not found.
            500: Error occurred while fetching device information from Keycloak.
        """
        session_data = []
        access_token = request.session.get("access_token")

        if not access_token:
            logger.warning("Access token not found for user: %s", request.user.username)
            return JsonResponse(
                {"error": True, "message": "Authentication required"}, status=401
            )

        try:
            session_data = get_active_devices(access_token)
            logger.info("Retrieved devices for user: %s", request.user.username)
            return JsonResponse({"session_data": session_data, "error": False})

        except KeycloakError as e:
            logger.error(
                "Failed to fetch active devices for user %s: %s",
                request.user.username,
                str(e),
            )
            return JsonResponse(
                {"error": True, "message": "Failed to retrieve device information"},
                status=500,
            )



from datetime import datetime

from django import template

register = template.Library()


@register.filter
def unix_to_datetime(value):
    try:
        return datetime.fromtimestamp(int(value))
    except (ValueError, TypeError):
        return ""


# This can be moved wherever it is more appropriate.
@register.filter
def in_any_group(user, group_names):
    """
    Template filter to check if a user belongs to any of the given groups.
    Usage: `if user|in_any_group:"group1,group2,group3`
    """
    if not user.is_authenticated:
        return False
    group_list = [group.strip() for group in group_names.split(",")]
    return user.groups.filter(name__in=group_list).exists()


@register.filter
def dynamic_logout_url(is_current):
    return "kc_auth_logout" if is_current else "kc_auth_remote-logout"
